# Copyright (c) Kuba Szczodrzyński 2022-05-12.

from svgwrite import Drawing

from ..mixins import HasVars, ParentType
from ..vector import V
from .base import Shape


class ShapeGroup(Shape, HasVars):
    name: str
    repeat: int = 1

    shapes: list[Shape] = []

    def __init__(self, core, parent: ParentType | None, *a, **kw):
        super().__init__(*a, **kw)

        if not self.name:  # used in wrap()
            return

        if isinstance(parent, HasVars):
            # apply parent variables
            vars = self.vars
            self.vars = dict(parent.vars)
            self.vars |= vars

        self.shapes = []
        self.vars["J"] = self.repeat
        for i in range(self.repeat):
            self.vars["I"] = i
            self.shapes += core.build_shapes(self.name, self, self.pos)
        self.vars.pop("I", None)
        self.vars.pop("J", None)

    def draw(self, dwg: Drawing):
        for shape in self.shapes:
            shape.draw(dwg)

    def get_by_id(self, id: str) -> Shape | None:
        for shape in self.shapes:
            if shape.fullid == id:
                return shape
            if isinstance(shape, ShapeGroup):
                result = shape.get_by_id(id)
                if result:
                    return result
        return None

    def get_by_id_path(self, path: str) -> Shape | None:
        path = path.split(".")
        id = path[0]
        for shape in self.shapes:
            if shape.base_id != id:
                continue
            if len(path) == 1:
                return shape
            if isinstance(shape, ShapeGroup):
                return shape.get_by_id_path(".".join(path[1:]))
            return shape
        return None

    @staticmethod
    def wrap(core, id: str, shapes: list[Shape]) -> "ShapeGroup":
        return ShapeGroup(
            core=core,
            parent=None,
            id=id,
            pos=V(0.0, 0.0),
            name="",
            shapes=shapes,
        )

    def move(self, vec: V):
        self.pos += vec
        for shape in self.shapes:
            shape.move(vec)

    @property
    def x1(self) -> float:
        return min(self.shapes, key=lambda t: t.x1).x1

    @property
    def y1(self) -> float:
        return min(self.shapes, key=lambda t: t.y1).y1

    @property
    def x2(self) -> float:
        return max(self.shapes, key=lambda t: t.x2).x2

    @property
    def y2(self) -> float:
        return max(self.shapes, key=lambda t: t.y2).y2



#  Copyright (c) Kuba Szczodrzyński 2023-6-3.

import wx
import wx.svg


class SvgPanel(wx.Control):
    Image: wx.svg.SVGimage | None = None

    def __init__(self, *args, **kw):
        super().__init__(*args, **kw)
        self.Bind(wx.EVT_PAINT, self.OnPaint)
        self.Bind(wx.EVT_SIZE, self.OnSize)
        self.Bind(wx.EVT_ERASE_BACKGROUND, self.OnEraseBackground)
        self.Bind(wx.EVT_PAINT, self.OnPaint)

    def ClearSvg(self) -> None:
        self.Image = None
        self.Refresh(False)

    def LoadSvg(self, svg: str) -> None:
        self.Image = wx.svg.SVGimage.CreateFromBytes(svg.encode())
        self.Refresh(False)

    def OnPaint(self, event) -> None:
        size = self.GetClientSize()
        dc = wx.BufferedPaintDC(self)
        dc.Clear()

        if not self.Image:
            return

        dcdim = min(size.width, size.height * 2)
        imgdim = max(self.Image.width, self.Image.height)
        scale = dcdim / imgdim
        int(self.Image.width * scale)
        int(self.Image.height * scale)

        r: wx.GraphicsRenderer = wx.GraphicsRenderer.GetDirect2DRenderer()
        ctx = r.CreateContext(dc)
        self.Image.RenderToGC(ctx, scale)

    def OnSize(self, event) -> None:
        self.Refresh(False)

    def OnEraseBackground(self, event) -> None:
        pass



# Copyright (c) Kuba Szczodrzyński 2022-05-11.

from . import models, shapes, utils
from .core import Core
from .mixins import HasId, HasVars, ParentType
from .readme import ReadmeWriter
from .vector import V

__all__ = [
    "models",
    "shapes",
    "utils",
    "Core",
    "HasVars",
    "HasId",
    "ParentType",
    "V",
    "ReadmeWriter",
]



# Copyright (c) Kuba Szczodrzyński 2022-05-11.

from enum import Enum


class ShapeType(Enum):
    RECT = "rect"
    CIRCLE = "circle"
    TEXT = "text"
    SUBSHAPE = "include"


class Side(Enum):
    FRONT = "front"
    BACK = "back"


class LabelDir(Enum):
    LEFT = "left"
    RIGHT = "right"


class RoleType(Enum):
    NC = "NC"
    C_NAME = "C_NAME"
    IO = "IO"
    IRQ = "IRQ"
    PHYSICAL = "PHYSICAL"
    PWR = "PWR"
    GND = "GND"
    CTRL = "CTRL"
    IC = "IC"
    GPIO = "GPIO"
    GPIONUM = "GPIONUM"
    ADC = "ADC"

    ARD_D = "ARD_D"
    ARD_A = "ARD_A"

    UART = "UART"
    I2C = "I2C"
    I2S = "I2S"
    DVP = "DVP"
    SPI = "SPI"
    SD = "SD"
    USB = "USB"
    IRDA = "IRDA"

    PWM = "PWM"
    TMR = "TMR"
    WAKE = "WAKE"
    RTC = "RTC"

    JTAG = "JTAG"
    SWD = "SWD"
    FLASH = "FLASH"


class IOType(Enum):
    NC = "NC"
    IO = "IO"
    I = "I"
    O = "O"
    PWM = "PWM"
    PWR = "PWR"
    NULL = "NULL"


RoleValue = list[str] | str | int | float | IOType | None



# Copyright (c) Kuba Szczodrzyński 2022-05-14.

import os
from os.path import dirname

from natsort import natsorted

from ..core import Core
from ..models import Board
from ..models.board import BoardDoc, BoardDocParams
from ..models.enums import RoleType
from ..models.flash_region import FlashRegion
from .parts import ReadmeParts


class ReadmeWriter(ReadmeParts):
    core: Core
    items: list[str] = []

    def __init__(self, core: Core) -> None:
        self.core = core
        self.items = []

    def clear(self):
        self.items = []

    def write(self, board: Board):
        if not board.doc:
            board.doc = BoardDoc()
        if not board.doc.params:
            board.doc.params = BoardDocParams()

        # Info
        self.add_heading(board.name)
        self.add_styled("*", "by", board.vendor)
        if board.url:
            self.add_link("Product page", board.url)
        if board.doc.links:
            self.add_list(
                *[[self.get_link(text, href)] for text, href in board.doc.links.items()]
            )

        # Specifications
        header = ["Parameter", "Value"]
        rows = [
            ["Board code", f"`{board.id}`"],
            ["MCU", board.build.mcu.upper()],
        ]
        if board.doc.params.manufacturer:
            rows.append(["Manufacturer", board.doc.params.manufacturer])
        if board.doc.params.series:
            rows.append(["Series", board.doc.params.series])
        rows.append(["Frequency", board.cpu_freq])
        rows.append(["Flash size", board.size_flash])
        rows.append(["RAM size", board.size_ram])
        if board.doc.params.voltage:
            rows.append(["Voltage", board.doc.params.voltage])
        if board.pcb and board.pcb.pinout:
            role_pins: dict[RoleType, set] = {}
            for pin in board.pcb.pinout.values():
                for role_type, roles in pin.items():
                    if role_type not in role_pins:
                        role_pins[role_type] = set()
                    if not isinstance(roles, list):
                        roles = [str(roles)]
                    for role in roles:
                        if role[0].isnumeric() and len(role) > 2 and role[1] == "_":
                            role_pins[role_type].add(role[0])
                        else:
                            role_pins[role_type].add(role)
            roles = [RoleType.GPIO, RoleType.PWM, RoleType.UART, RoleType.ADC]
            counts = [
                f"{len(role_pins[r])}x {r.name}"
                for r in roles
                if r in role_pins and len(role_pins[r])
            ]
            rows.append(["I/O", ", ".join(counts)])
        if board.doc.params.extra:
            rows.extend([[k, v] for k, v in board.doc.params.extra.items()])
        if board.doc.fccid:
            link = self.get_link(board.doc.fccid, f"https://fccid.io/{board.doc.fccid}")
            rows.append(["FCC ID", link])
        self.add_table(header, *rows)

        # Usage
        self.add_heading("Usage", 2)
        self.add_text("**Board code:**", f"`{board.id}`")
        if self.core.is_libretiny:
            family_component_map = {
                "realtek-amb": "rtl87xx",
                "beken-72xx": "bk72xx",
            }

            self.add_text("In `platformio.ini`:")
            code = [
                f"[env:{board.id}]",
                "platform = libretiny",
                f"board = {board.id}",
                "framework = arduino",
            ]
            self.add_code(code, lang="ini")

            component = None
            # noinspection PyBroadException
            try:
                from ltchiptool import Family

                family = Family.get(board.build.family)
                for f in family.inheritance:
                    if f.name in family_component_map:
                        component = family_component_map[f.name]
                        break
            except Exception:
                pass

            if component:
                self.add_text("In ESPHome YAML:")
                code = [
                    f"{component}:",
                    "  board: " + board.id,
                ]
                self.add_code(code, lang="yaml")

        # Pinout
        if board.pcb and board.pcb.pinout:
            if board.pcb.templates:
                self.add_heading("Pinout", 2)
                self.add_img("Pinout", f"{board.id}.svg")

            self.add_heading("Pin functions", 2)
            header = ["Name(s)", "UART", "I²C", "SPI", "PWM", "Other"]
            rows = []
            gpio_pins: dict[str, dict[RoleType, list[str]]] = {}
            roles = self.core.roles
            hidden = board.pcb.pinout_hidden.split(",")

            for pin in board.pcb.pinout.values():
                if RoleType.GPIO not in pin:
                    continue
                gpio = pin[RoleType.GPIO]
                gpio_pins[gpio] = {}
                for role_type, functions in pin.items():
                    if role_type not in roles:
                        continue
                    role_text = roles[role_type].format(
                        functions,
                        long=False,
                        hidden=hidden,
                    )
                    gpio_pins[gpio][role_type] = role_text

            for num in natsorted(gpio_pins.keys()):
                pin = gpio_pins[num]
                rows.append(
                    [
                        pin.get(RoleType.GPIO, []) + pin.get(RoleType.ADC, []),
                        pin.get(RoleType.UART, []),
                        pin.get(RoleType.I2C, []),
                        pin.get(RoleType.SPI, []),
                        pin.get(RoleType.PWM, []),
                        pin.get(RoleType.SWD, [])
                        + pin.get(RoleType.JTAG, [])
                        + pin.get(RoleType.DVP, []),
                    ]
                )
            for i, row in enumerate(rows):
                rows[i] = [
                    ", ".join(col) if isinstance(col, list) else col for col in row
                ]
            self.add_table(header, *rows)

        # Flash
        if board.flash:
            self.add_heading("Flash memory map", 2)
            self.add_text(
                "Flash size:",
                board.size_flash,
                "/",
                f"{format(board.upload.flash_size, ',d')} B",
                "/",
                "0x%X" % board.upload.flash_size,
            )
            self.add_text("Hex values are in bytes.")
            names = self.core.flash
            flash = sorted(board.flash, key=lambda reg: reg.start)
            header = ["Name", "Start", "Length", "End"]
            rows = []
            prev_end = 0
            for reg in flash:
                reg.hex_size_len = 0
                name = names[reg.name] if reg.name in names else reg.name
                if prev_end < reg.start:
                    res = FlashRegion(
                        name="res",
                        offset=prev_end,
                        length=reg.start - prev_end,
                    )
                    res.hex_offs_len = reg.hex_offs_len
                    rows.append(["(reserved)"] + res.lst)
                rows.append([name] + reg.lst)
                prev_end = reg.end
            self.add_table(header, *rows)

        # Extras
        if board.doc.extra:
            self.items += board.doc.extra

    def save(self, output: str):
        os.makedirs(dirname(output), exist_ok=True)
        with open(output, "w", encoding="utf-8") as f:
            f.write("\n\n".join(self.items))
            f.write("\n")



# Copyright (c) Kuba Szczodrzyński 2022-05-12.

from svgwrite import Drawing, text

from ..utils import EvalFloat
from .base import Shape
from .fill_style import FillStyle


class Text(Shape):
    text: str
    font_size: EvalFloat
    fill: FillStyle = None

    def draw(self, dwg: Drawing):
        txt = text.Text(
            text=self.text,
            insert=self.pos.tuple,
            id=self.fullid,
            font_family="Consolas",
            font_size=self.font_size,
        )
        if self.fill:
            self.fill.apply_to(dwg, txt)
        dwg.add(txt)

    @property
    def x1(self) -> float:
        return self.pos.x - 1

    @property
    def y1(self) -> float:
        return self.pos.y - 1

    @property
    def x2(self) -> float:
        return self.pos.x + 1

    @property
    def y2(self) -> float:
        return self.pos.y + 1



#  Copyright (c) Kuba Szczodrzyński 2023-6-3.

from svgwrite import Drawing, shapes
from svgwrite.utils import AutoID

from .core import Core
from .models import Pcb, Side
from .shapes import Shape, ShapeGroup
from .vector import V


def get_pcb_images(core: Core, pcb: Pcb, with_labels: bool) -> list[Shape]:
    shapes = []
    for side in [Side.FRONT, Side.BACK]:
        shape = pcb.shapes[side]
        if with_labels:
            labels, _, _ = core.build_labels(pcb, side)
            if labels:
                shapes.append(ShapeGroup.wrap(core, side.value, [shape] + labels))
        else:
            shapes.append(shape)
    return shapes


def draw_shapes(
    px_size: V, scale: float | None, images: list[Shape], with_canvas: bool
) -> Drawing:
    AutoID._set_value(1)
    dwg = Drawing(size=px_size.tuple)

    # stack horizontally
    part_size = V(px_size.x / len(images), px_size.y)
    part_pos = [V(part_size.x * i, 0) for i in range(len(images))]

    if scale:
        vb_size = px_size / scale
    else:
        scale = 99999
        size_pad = part_size * 0.90  # 5% padding from each side
        for shape in images:
            size = shape.size
            scale = min(scale, size_pad.x / size.x, size_pad.y / size.y)
        print(" - calculated scale: %.2f" % scale)
        vb_size = px_size / scale

    dwg.viewbox(width=vb_size.x, height=vb_size.y)

    if with_canvas:
        bg = shapes.Rect(insert=(0, 0), size=vb_size.tuple)
        bg.fill(color="white")
        bg.stroke(color="black", width=0.1)
        dwg.add(bg)

    for i, shape in enumerate(images):
        shape_pos = ((part_size / scale) - shape.size) / 2
        shape_pos += part_pos[i] / scale
        shape_pos -= shape.pos1
        if with_canvas:
            shape_pos.x -= 0.05
        shape.move(shape_pos)
        shape.draw(dwg)
        shape.move(-shape_pos)

    return dwg



#  Copyright (c) Kuba Szczodrzyński 2023-5-20.

from typing import Any, Dict

from ltctplugin.base import PluginBase


class Plugin(PluginBase):
    @property
    def title(self) -> str:
        return "boardgen"

    @property
    def has_gui(self) -> bool:
        return True

    def build_gui(self, *args, **kwargs) -> Dict[str, Any]:
        from .gui import BoardgenPanel

        return dict(
            boardgen=BoardgenPanel,
        )


entrypoint = Plugin

__all__ = [
    "entrypoint",
]



# Copyright (c) Kuba Szczodrzyński 2022-06-15.

import os
import re
from os.path import dirname

from ..core import Core
from ..models.board import Board
from ..models.enums import RoleType
from ..models.pcb import PinDict
from .features import PinFeatures
from .parts import VariantParts
from .section import SectionType

# map of RoleTypes to variant.h SectionTypes
SECTION_MAP = {
    RoleType.SPI: SectionType.SPI,
    RoleType.I2C: SectionType.WIRE,
    RoleType.UART: SectionType.SERIAL,
}

# map of RoleTypes to PinFeatures
FEATURE_MAP = {
    RoleType.ARD_D: PinFeatures.PIN_GPIO,
    RoleType.ARD_A: PinFeatures.PIN_ADC,
    RoleType.PWM: PinFeatures.PIN_PWM,
    RoleType.I2C: PinFeatures.PIN_I2C,
    RoleType.I2S: PinFeatures.PIN_I2S,
    RoleType.IRQ: PinFeatures.PIN_IRQ,
    RoleType.JTAG: PinFeatures.PIN_JTAG,
    RoleType.SPI: PinFeatures.PIN_SPI,
    RoleType.SWD: PinFeatures.PIN_SWD,
    RoleType.UART: PinFeatures.PIN_UART,
}

# (indexed) communication ports with lists of required signals
PORT_SIGNALS = {
    RoleType.SPI: ["SCK", "MISO", "MOSI"],
    RoleType.I2C: ["SDA", "SCL"],
    RoleType.UART: ["TX"],
}

# list of RoleTypes to include in SectionType.MACROS
MACROS_ROLES = list(PORT_SIGNALS.keys()) + [
    RoleType.GPIO,
    RoleType.ADC,
    RoleType.PWM,
]

# RoleTypes not included in the variant.cpp comment lines
ROLES_HIDDEN = ["ARD_A", "ARD_D", "IO", "C_NAME", "PHYSICAL"]


class VariantWriter(VariantParts):
    core: Core

    def __init__(self, core: Core) -> None:
        self.core = core
        self.pins = {}
        self.sections = {}
        self.gpio_map = {}
        self.static_pins = {}

    @staticmethod
    def read_pin(pin: PinDict) -> tuple[str, str, int] | None:
        name = pin.get(RoleType.GPIO, None) or pin.get(RoleType.ADC, None)
        c_name = pin.get(RoleType.C_NAME, name)
        if not name:
            return None
        number = pin.get(RoleType.GPIONUM, None)
        if number is not None:
            return name, c_name, int(number)
        number = re.sub(r"\D", "", name)
        return name, c_name, int(number)

    def generate(self, board: Board):
        pcb = board.pcb
        if not pcb or not pcb.pinout:
            return
        if not board.has_arduino_core:
            return

        self.add_item(SectionType.PINS, "PINS_COUNT", 0, "Total GPIO count")
        self.add_item(SectionType.PINS, "NUM_DIGITAL_PINS", 0, "Digital inputs/outputs")
        self.add_item(SectionType.PINS, "NUM_ANALOG_INPUTS", 0, "ADC inputs")
        self.add_item(SectionType.PINS, "NUM_ANALOG_OUTPUTS", 0, "PWM & DAC outputs")

        # { role_type: { index: { signal: [...pin_numbers] } } }
        ports: dict[RoleType, dict[int, dict[str, list[int]]]] = {}
        for role_type in PORT_SIGNALS.keys():
            ports[role_type] = {}
        max_pin_number = 0

        for pin in pcb.pinout.values():
            pin_tuple = self.read_pin(pin)
            if not pin_tuple:
                continue
            pin_name, c_name, pin_number = pin_tuple

            arduino_name = None
            ard_d = pin.get(RoleType.ARD_D, None)
            ard_a = pin.get(RoleType.ARD_A, None)
            if ard_d:
                self.add_item(SectionType.ARDUINO, f"PIN_{ard_d}", pin_number, c_name)
                self.static_pins[ard_d] = f"PIN_{ard_d}"
                arduino_name = ard_d
            if ard_a:
                self.add_item(SectionType.ARDUINO, f"PIN_{ard_a}", pin_number, c_name)
                self.static_pins[ard_a] = f"PIN_{ard_a}"
                arduino_name = arduino_name or ard_a

            if not arduino_name:
                continue
            self.gpio_map[pin_number] = arduino_name
            max_pin_number = max(max_pin_number, pin_number)

            pin_comment = []
            for role_type, values in pin.items():
                if role_type.name in ROLES_HIDDEN:
                    continue
                role = self.core.role(role_type)
                if not role:
                    continue
                pin_comment += role.format(values, long=True, hidden=ROLES_HIDDEN)
                if role_type not in MACROS_ROLES:
                    continue
                roles_short = role.format(values, long=False, hidden=ROLES_HIDDEN)
                for text in roles_short:
                    self.add_item(
                        section_type=SectionType.MACROS,
                        key=f"PIN_{text}",
                        value=pin_number,
                        comment=c_name,
                        single=True,
                    )
            pin_comment = ", ".join(pin_comment)

            # add the pin to the list
            new_added = self.add_pin(arduino_name, c_name, pin_comment)
            # calculate new counters
            if new_added:
                self.increment_item(SectionType.PINS, "PINS_COUNT")
                if RoleType.ARD_D in pin:
                    self.increment_item(SectionType.PINS, "NUM_DIGITAL_PINS")
                if RoleType.ARD_A in pin:
                    self.increment_item(SectionType.PINS, "NUM_ANALOG_INPUTS")
                if RoleType.PWM in pin:
                    self.increment_item(SectionType.PINS, "NUM_ANALOG_OUTPUTS")

            # add roles for the pin
            for role_type in pin.keys():
                if role_type not in FEATURE_MAP:
                    continue
                self.add_pin_feature(arduino_name, FEATURE_MAP[role_type])

            # find all indexed interfaces this pin belongs to
            for role_type, roles in pin.items():
                if role_type not in ports:
                    continue
                if not isinstance(roles, list):
                    roles = [str(roles)]
                for role in roles:
                    # require "1_RX" format
                    if not role[0].isnumeric() or role[1] != "_":
                        continue
                    (index, _, port) = role.partition("_")
                    index = int(index)
                    if index not in ports[role_type]:
                        ports[role_type][index] = {}
                    if port not in ports[role_type][index]:
                        ports[role_type][index][port] = []
                    ports[role_type][index][port].append(pin_number)

        # { role_type: { index: { signal: [...pin_numbers] } } }
        for role_type, interfaces in ports.items():
            section = SECTION_MAP[role_type]
            for index, signals in interfaces.items():
                if not all(s in signals for s in PORT_SIGNALS[role_type]):
                    # skip incomplete ports (i.e. only SDA1 available)
                    continue
                self.increment_item(
                    SectionType.PORTS,
                    f"{section.name}_INTERFACES_COUNT",
                )
                for signal, pin_numbers in signals.items():
                    self.add_item(SectionType.PORTS, f"HAS_{section.name}{index}", 1)
                    key = f"PIN_{section.name}{index}_{signal}"
                    if len(pin_numbers) > 1:
                        for i, pin in enumerate(pin_numbers):
                            c_name = self.pins[self.gpio_map[pin]][0]
                            self.add_item(section, f"{key}_{i}", pin, c_name)
                    else:
                        c_name = self.pins[self.gpio_map[pin_numbers[0]]][0]
                        self.add_item(section, key, pin_numbers[0], c_name)
                    key = f"PINS_{section.name}{index}_{signal}"
                    pins_array = ", ".join(f"{pin}u" for pin in pin_numbers)
                    self.add_item(section, key, f"(pin_size_t[]){{{pins_array}}}")

        self.add_item(
            SectionType.PINS,
            "PINS_GPIO_MAX",
            max_pin_number,
            "Last usable GPIO number",
        )

    def save_h(self, output: str, board_name: str):
        os.makedirs(dirname(output), exist_ok=True)
        with open(output, "w", encoding="utf-8") as f:
            lines = [
                f"/* This file was auto-generated from {board_name} using boardgen */",
                "",
                "#pragma once",
                "",
                self.format_sections(),
            ]
            f.write("\n".join(lines))

    def save_c(self, output: str, board_name: str):
        os.makedirs(dirname(output), exist_ok=True)
        with open(output, "w", encoding="utf-8") as f:
            lines = [
                f"/* This file was auto-generated from {board_name} using boardgen */",
                "",
                "#include <Arduino.h>",
                "",
                "#ifdef LT_VARIANT_INCLUDE",
                "#include LT_VARIANT_INCLUDE",
                "#endif",
                "",
                self.format_pins(),
                "",
            ]
            f.write("\n".join(lines))



# Copyright (c) Kuba Szczodrzyński 2022-05-14.

from abc import ABC


class ReadmeParts(ABC):
    items: list[str]

    def pad(self, s: str, i: int) -> str:
        return s + " " * (i - len(s))

    def add_heading(self, text: str, level: int = 1) -> "ReadmeParts":
        self.items.append(level * "#" + " " + text)
        return self

    def get_link(self, text: str, href: str) -> str:
        return f"[{text}]({href})"

    def get_img(self, alt: str, src: str) -> str:
        return f"![{alt}]({src})"

    def add_link(self, text: str, href: str) -> "ReadmeParts":
        self.items.append(self.get_link(text, href))
        return self

    def add_img(self, alt: str, src: str) -> "ReadmeParts":
        self.items.append(self.get_img(alt, src))
        return self

    def add_text(self, *text: str) -> "ReadmeParts":
        self.items.append(" ".join(text))
        return self

    def add_styled(self, style: str, *text: str) -> "ReadmeParts":
        self.items.append(style + " ".join(text) + style)
        return self

    def add_list(self, *items: list[str]) -> "ReadmeParts":
        items = [" ".join(i) for i in items]
        self.items.append("- " + "\n- ".join(items))
        return self

    def add_code(self, code: str | list[str], lang: str = None) -> "ReadmeParts":
        if isinstance(code, list):
            code = "\n".join(code)
        if not lang:
            lang = ""
        self.items.append(f"```{lang}\n{code}\n```")

    def add_table(self, header: list[str], *rows: list[str]) -> "ReadmeParts":
        maxlen = [len(h) for h in header]
        for row in rows:
            for i, col in enumerate(row):
                maxlen[i] = max(maxlen[i], len(col))
        lines = []
        header = [self.pad(h, maxlen[i]) for i, h in enumerate(header)]
        line = " | ".join(header)
        lines.append(line.rstrip())
        underline = ["-" * i for i in maxlen]
        line = "-|-".join(underline)
        lines.append(line.rstrip())
        for row in rows:
            row = [self.pad(h, maxlen[i]) for i, h in enumerate(row)]
            line = " | ".join(row)
            lines.append(line.rstrip())
        self.items.append("\n".join(lines))
        return self



# Copyright (c) Kuba Szczodrzyński 2022-05-12.

from svgwrite import Drawing

from ...models.enums import IOType, RoleType, RoleValue
from ...models.role import Role
from ...shapes.base import Shape
from ...vector import V
from ..base import LabelShape
from .block import Block
from .io_line import IOLine


class Label(LabelShape):
    roles: dict[RoleType, RoleValue]
    labels: list[LabelShape] = []

    def build(
        self,
        core,
        pin: str,
        pad: Shape,
        hidden: list[str],
        block_extra: dict = {},
        io_extra: dict = {},
    ):
        # not connected pin
        if RoleType.NC in self.roles:
            self.roles[RoleType.IO] = IOType.NC.value
        # PWM pin
        if RoleType.PWM in self.roles:
            self.roles[RoleType.IO] = IOType.PWM.value
        # GPIO pin - set I/O role
        if RoleType.GPIO in self.roles and RoleType.IO not in self.roles:
            self.roles[RoleType.IO] = IOType.IO.value
        # power or ground pin
        if RoleType.PWR in self.roles or RoleType.GND in self.roles:
            self.roles[RoleType.IO] = IOType.PWR.value
        # ground text
        if RoleType.GND in self.roles:
            self.roles[RoleType.GND] = "GND"
        # no IO role
        if RoleType.IO not in self.roles:
            self.roles[RoleType.IO] = IOType.NULL.value

        # sort labels according to enum sorting
        role_types = list(RoleType)
        roles = sorted(self.roles.items(), key=lambda x: role_types.index(x[0]))

        # build label shapes
        pos = V(self.pos)
        for role_type, functions in roles:
            if role_type not in core.roles:
                continue
            role: Role = core.roles[role_type]
            texts = role.format(functions, hidden=hidden)
            for text in texts:
                params = dict(
                    pos=V(pos),
                    label_dir=pad.label_dir,
                    label_size=pad.label_size,
                    role_type=role_type,
                    ratio=role.ratio,
                    color=role.color,
                )
                match role_type:
                    case RoleType.IO:
                        shape = IOLine(
                            **params,
                            type=IOType(text),
                            **io_extra,
                        )
                    case _:
                        shape = Block(
                            **params,
                            text=text,
                            **block_extra,
                        )
                pos.x += (shape.width + shape.padding.x * 2) * shape.dirv
                self.labels.append(shape)

    def move(self, vec: V):
        self.pos += vec
        for shape in self.labels:
            shape.move(vec)

    def draw(self, dwg: Drawing):
        for shape in self.labels:
            shape.draw(dwg)

    @property
    def x1(self) -> float:
        return min(self.labels, key=lambda t: t.x1).x1

    @property
    def y1(self) -> float:
        return min(self.labels, key=lambda t: t.y1).y1

    @property
    def x2(self) -> float:
        return max(self.labels, key=lambda t: t.x2).x2

    @property
    def y2(self) -> float:
        return max(self.labels, key=lambda t: t.y2).y2



#  Copyright (c) Kuba Szczodrzyński 2023-6-3.

import json
from copy import deepcopy
from enum import Enum, auto
from logging import debug, info, warning
from os import makedirs, rename, unlink
from os.path import abspath, basename, dirname, join
from shutil import copyfile

import wx
import wx.adv
import wx.xrc
from ltchiptool import Family
from ltchiptool.gui.panels.base import BasePanel
from ltchiptool.gui.utils import on_event, with_event
from ltchiptool.util.lvm import LVM

from boardgen import Core, HasVars, V
from boardgen.draw_util import draw_shapes, get_pcb_images
from boardgen.models import Board, RoleType, Template
from boardgen.shapes import Shape, ShapeGroup

from .svg import SvgPanel
from .utils import jsonpath, jsonwalk

INIT_BOARD = {
    "_base": [],
    "build": {
        "mcu": "",
        "variant": "",
    },
    "name": "",
    "url": "",
    "vendor": "",
    "doc": {
        "fccid": "",
    },
    "pcb": {
        "symbol": "",
    },
}
INIT_TEMPLATE = {
    "name": "",
    "title": "",
    "width": 10,
    "height": 10,
    "front": [],
    "back": [],
    "pads": {},
    "test_pads": {},
}
INIT_SHAPE = []
ROLE_DEFAULTS: dict[str, str | int | None] = {
    "ADC": 0,
    "ARD": "D0",
    "CTRL": "CEN",
    "C_NAME": "GPIO_00",
    "DVP": "PD0",
    "FLASH": "^FCS",
    "GND": None,
    "GPIO": "P0",
    "GPIONUM": 35,
    "I2C": "0_SDA",
    "I2S": "0_TX",
    "IC": 1,
    "IO": "I",
    "IRDA": None,
    "IRQ": None,
    "JTAG": "TMS",
    "PHYSICAL": 2,
    "PWM": 1,
    "PWR": 3.3,
    "SD": "CMD",
    "SPI": "0_SCK",
    "SWD": "DIO",
    "UART": "1_TX",
    "USB": "DN",
    "WAKE": 1,
}
SHAPE_DEFAULTS = {
    "circle": {
        "type": "circle",
        "pos": "0,0",
        "d": 1.0,
        "fill": {
            "color": "black",
        },
    },
    "rect": {
        "type": "rect",
        "pos": "0,0",
        "size": "1,1",
        "fill": {
            "color": "black",
        },
    },
    "subshape": {
        "id": "subshape_id_here",
        "name": "test_pad_1mm",
        "repeat": 1,
        "pos": "0,0",
    },
    "text": {
        "type": "text",
        "pos": "0,0",
        "text": "Text",
        "font_size": 1.0,
        "fill": {
            "color": "red",
        },
    },
}


class EditType(Enum):
    # multiple choice
    BASE = auto()
    TEMPLATE = auto()
    CONNECTIVITY = auto()
    ROLE_HIDDEN = auto()
    # single choice
    SHAPE = auto()
    SHAPE_ADD = auto()
    PRESET = auto()
    FAMILY = auto()
    ROLE = auto()
    FLASH = auto()
    # other
    COLOR = auto()


class BoardgenPanel(BasePanel):
    file_map: dict[str, str] = None
    draw_items: dict[str, str] = None
    draw_object: Board | list[Shape] | Shape
    edit_items: dict[str, tuple[str, dict]] = None
    edit_selection: str = None
    edit_path: str = ""
    edit_errors: bool = False
    edit_type: EditType | None = None
    modified: dict[str, dict] = None
    _vars: dict

    def __init__(self, parent: wx.Window, frame):
        super().__init__(parent, frame)
        self.LoadXRCFile("boardgen.xrc")

        self._shapes = []
        self._vars = {}

        self.Sizer = wx.BoxSizer(wx.VERTICAL)
        self.Splitter = wx.SplitterWindow(self, style=wx.SP_3D | wx.SP_LIVE_UPDATE)
        self.Left: wx.Panel = self.Xrc.LoadPanel(self.Splitter, "PreviewPanel")
        self.Right: wx.Panel = self.Xrc.LoadPanel(self.Splitter, "EditPanel")
        self.Splitter.SetMinimumPaneSize(300)
        self.Splitter.SetSashGravity(0.3)
        self.Splitter.SplitVertically(self.Left, self.Right, sashPosition=-600)
        self.Sizer.Add(self.Splitter, proportion=1, flag=wx.EXPAND)
        self.SetSizer(self.Sizer)

        self.AddToNotebook("boardgen")

        self.PreviewPanel: wx.Panel = self.Left.FindWindowByName("preview_panel", self)
        self.PreviewBox: wx.BoxSizer = self.PreviewPanel.GetSizer()
        self.Svg = SvgPanel(self.PreviewPanel)
        self.PreviewBox.Add(self.Svg, proportion=1, flag=wx.EXPAND)

        self.EditItem = self.BindComboBox("combo_edit_item")
        self.DrawItem = self.BindComboBox("combo_preview_item")
        self.Data: wx.TextCtrl = self.FindWindowByName("input_data", self)
        self.Data.Bind(wx.EVT_TEXT, self.OnDataText)
        self.Data.Bind(wx.EVT_KEY_UP, self.OnDataPosition)
        self.Data.Bind(wx.EVT_LEFT_UP, self.OnDataPosition)
        self.Vars = self.BindTextCtrl("input_vars")
        self.Vars.Bind(wx.EVT_TEXT, self.OnVarsText)
        self.Path = self.BindTextCtrl("input_path")
        self.Error = self.BindTextCtrl("input_error")
        self.Save = self.BindButton("button_save", self.OnSaveClick)
        self.Edit = self.BindButton("button_edit", self.OnEditClick)
        self.Format = self.BindButton("button_format", self.OnFormatClick)
        self.Choose = self.BindButton("button_choose", self.OnChooseClick)
        self.Revert = self.BindButton("button_revert", self.OnRevertClick)
        self.Discard = self.BindButton("button_discard", self.OnDiscardClick)
        self.Modified: wx.adv.HyperlinkCtrl = self.BindWindow(
            "label_modified",
            (wx.adv.EVT_HYPERLINK, self.OnModifiedClick),
        )

        self.lvm = LVM.get()
        self.core = Core()
        self.core.add_custom_dirs(
            boards=join(self.lvm.path(), "boards"),
            shapes=join(self.lvm.path(), "boards", "shapes"),
            templates=join(self.lvm.path(), "boards", "templates"),
        )
        self.core.json_hook = self.AddEditItem

        self.file_map = {}
        self.draw_items = {}
        self.edit_items = {}
        self.modified = {}

        self.ReloadLists()

    def GetSettings(self) -> dict:
        return dict(
            draw_item=self.draw_item,
            edit_item=self.edit_item and self.edit_item[0],
            split=self.Splitter.GetSashPosition(),
            user_vars=self.user_vars,
        )

    def SetSettings(
        self,
        draw_item: str = None,
        edit_item: str = None,
        split: int = None,
        user_vars: dict[str, str] = None,
        **_,
    ) -> None:
        if draw_item:
            self.draw_item = draw_item
        if edit_item:
            self.edit_item = edit_item
        if split:
            self.Splitter.SetSashPosition(split)
        if user_vars:
            self.user_vars = user_vars

    def ReloadLists(self) -> None:
        self.core.clear_cache()
        draw_selection = self.DrawItem.GetStringSelection()
        items = {}

        for board_name in self.core.list_json("boards"):
            items[f"Board - {board_name}"] = f"boards/{board_name}"
        for template_name in self.core.list_json("templates"):
            items[f"Template - {template_name}"] = f"templates/{template_name}"
        for shape_name in self.core.list_json("shapes"):
            items[f"Shape - {shape_name}"] = f"shapes/{shape_name}"

        self.DrawItem.SetItems(sorted(items.keys()))
        self.DrawItem.SetStringSelection(draw_selection)
        self.draw_items = items

    def OnUpdate(self, target: wx.Window = None) -> None:
        super().OnUpdate(target)
        if not self.IsShown():
            return
        match target:
            case self.DrawItem:
                self.Svg.ClearSvg()
                self.UpdateDrawItem()
            case self.EditItem:
                self.UpdateEditItem()

            case _:
                # debug(f"OnUpdate({target})")
                pass

    def UpdateDrawItem(self, force_list: bool = True) -> None:
        item = self.draw_item
        if not item:
            return
        item_type, _, item_name = item.partition("/")

        self.edit_items = {}
        obj: dict = {}
        try:
            match item_type:
                case "boards":
                    obj = self.core.load_board(item_name, allow_cache=False)
                case "templates":
                    obj = self.core.load_template(item_name)
                case "shapes":
                    obj = self.core.load_shape(item_name)
        except Exception as e:
            self.SetError(e)
            if force_list:
                self.FillEditList()
            return
        self.vars = obj["vars"] if "vars" in obj else {}
        parent = HasVars(vars=self.vars)

        try:
            match item_type:
                case "boards":
                    self.draw_object = self.core.get_board(item_name)

                case "templates":
                    front: list[Shape] = []
                    back: list[Shape] = []
                    template = Template(**obj)
                    for data in template.front:
                        shape = self.core.build_shape(parent=parent, data=data)
                        front.append(shape)
                    for data in template.back:
                        shape = self.core.build_shape(parent=parent, data=data)
                        back.append(shape)
                    if front and back:
                        self.draw_object = [
                            ShapeGroup.wrap(self.core, item_name, front),
                            ShapeGroup.wrap(self.core, item_name, back),
                        ]
                    elif front:
                        self.draw_object = ShapeGroup.wrap(self.core, item_name, front)
                    elif back:
                        self.draw_object = ShapeGroup.wrap(self.core, item_name, back)

                case "shapes":
                    shapes: list[Shape] = []
                    for data in obj:
                        shape = self.core.build_shape(parent=parent, data=data)
                        shapes.append(shape)
                    self.draw_object = ShapeGroup.wrap(self.core, item_name, shapes)

            self.Redraw()
            self.SetError(None)
        except Exception as e:
            self.SetError(e)
            if not force_list:
                return
        self.FillEditList()

    def FillEditList(self) -> None:
        selection = self.EditItem.GetStringSelection()
        self.EditItem.Clear()
        self.EditItem.Append(list(self.edit_items.keys()))
        self.EditItem.Enable(True)
        self.EditItem.SetStringSelection(selection)

        if not self.EditItem.GetStringSelection():
            if self.EditItem.GetItems():
                self.edit_item = self.EditItem.GetItems()[0]
            else:
                self.Data.Clear()
                self.Path.Clear()

    def AddEditItem(self, type: str, name: str, data: dict, file: str | None) -> None:
        key = f"{type} -> {name}".replace("\\", "/")
        value = f"{type}/{name}", data
        self.edit_items[key] = value
        if file:
            file = abspath(file)
            self.file_map[value[0]] = file

    def UpdateEditItem(self) -> None:
        item = self.edit_item
        if not item:
            return
        item_name, obj = item
        self.edit_selection = item_name
        scroll_pos = self.Data.GetScrollPos(wx.VERTICAL)
        cursor_pos = self.Data.GetInsertionPoint()
        self.Data.ChangeValue(json.dumps(obj, indent=4))
        self.Data.SetInsertionPoint(cursor_pos)
        self.Data.SetScrollPos(wx.VERTICAL, scroll_pos)
        self.Format.Enable(True)
        self.Revert.Enable(item_name in self.modified)
        self.UpdateDataPosition()
        if "pcb" in obj:
            obj = obj["pcb"]
        self.vars = obj["vars"] if "vars" in obj else {}

    def Redraw(self) -> None:
        debug(f"Draw object: {self.draw_object}")
        match self.draw_object:
            case Board():
                images = get_pcb_images(
                    self.core,
                    self.draw_object.pcb,
                    with_labels=True,
                )
            case list():
                images = self.draw_object
            case Shape():
                images = [self.draw_object]
            case _:
                return
        if not images:
            self.Svg.ClearSvg()
            return
        dwg = draw_shapes(V(1024, 500), 12, images, with_canvas=False)
        self.Svg.LoadSvg(dwg.tostring())

    @with_event
    def OnDataText(self, event: wx.Event) -> None:
        event.Skip()
        try:
            text = self.Data.GetValue()
            obj = json.loads(text)
            if obj:
                self.data = obj
                if self.edit_item:
                    self.MarkModified()
            self.edit_errors = False
            self.UpdateDrawItem(force_list=self.edit_path.startswith("_base"))
        except json.JSONDecodeError:
            self.edit_errors = True

    @with_event
    def OnVarsText(self, event: wx.Event) -> None:
        event.Skip()
        self.UpdateDrawItem()

    @with_event
    def OnDataPosition(self, event: wx.Event) -> None:
        event.Skip()
        self.UpdateDataPosition()

    def UpdateDataPosition(self) -> None:
        text = self.Data.GetValue()
        text = text.replace("\n", "\r\n")
        pos = max(0, self.Data.GetInsertionPoint() - 1)
        self.edit_path = jsonpath(text, pos) or ""
        if self.edit_path:
            self.Path.ChangeValue(self.edit_path.replace(".", " -> "))
        else:
            self.Path.ChangeValue("")
        self.UpdateEditType()

    @on_event
    def OnSaveClick(self) -> None:
        self.Modified.SetLabel("Modified: 0")
        self.Save.Enable(False)
        self.Discard.Enable(False)
        self.Revert.Enable(False)
        if not self.modified:
            return
        for file, obj in self.modified.items():
            path = self.file_map.get(file, None)
            if not path:
                warning(f"Couldn't save '{file}' - path is unknown")
                continue
            with open(path, "w", encoding="utf-8") as f:
                json.dump(obj, f, indent="\t")
                f.write("\n")
        info(f"Saved {len(self.modified)} files")
        self.modified.clear()

    @on_event
    def OnDiscardClick(self) -> None:
        self.Modified.SetLabel("Modified: 0")
        self.Save.Enable(False)
        self.Discard.Enable(False)
        self.Revert.Enable(False)
        if not self.modified:
            return
        self.modified.clear()
        self.core.clear_cache()
        self.UpdateDrawItem()
        self.UpdateEditItem()

    @on_event
    def OnRevertClick(self) -> None:
        self.MarkUnmodified()
        name, _ = self.edit_item
        item_type, _, item_name = name.partition("/")
        self.core.remove_from_cache(item_type, item_name)
        self.UpdateDrawItem()
        self.UpdateEditItem()

    @on_event
    def OnModifiedClick(self) -> None:
        if not self.modified:
            message = "No files have been modified"
        else:
            message = "The following files have been modified:\n\n"
            for file in sorted(self.modified.keys()):
                path = self.file_map.get(file, file)
                message += path + "\n"
        wx.MessageBox(message, "Information")

    @on_event
    def OnFormatClick(self) -> None:
        if self.edit_errors:
            wx.MessageBox(
                "Can't beautify, the JSON has syntax errors",
                "Error",
                wx.OK | wx.CENTRE | wx.ICON_ERROR,
            )
            return
        self.UpdateEditItem()

    @on_event
    def OnEditClick(self) -> None:
        bar: wx.MenuBar = self.Xrc.LoadMenuBar("EditMenuBar")
        menu: wx.Menu = bar.GetMenu(0)
        menu.Detach()
        if not self.draw_item:
            menu.FindItemByPosition(1).Enable(False)
            menu.FindItemByPosition(2).Enable(False)
            menu.FindItemByPosition(3).Enable(False)
        choice = self.Edit.GetPopupMenuSelectionFromUser(menu)
        item: wx.MenuItem = menu.FindItemById(choice)
        if not item:
            return
        label = item.GetItemLabel()
        parent_label = None

        owner_menu: wx.Menu = item.GetMenu()
        owner_menu_parent: wx.Menu = owner_menu.GetParent()
        if owner_menu_parent:
            for item in owner_menu_parent.GetMenuItems():
                if item.GetSubMenu() == owner_menu:
                    parent_label = item.GetItemLabel()
                    break

        match parent_label, label:
            case "Board...", "LibreTiny directory":
                self.CreateNewFile("boards", self.lvm.path(), INIT_BOARD)
            case "Board...", "boardgen directory":
                self.CreateNewFile("boards", self.core.dir_base, INIT_BOARD)

            case "Board base...", "LibreTiny directory":
                new_name = self.AskFileName()
                new_path = join(
                    self.lvm.path(),
                    "boards",
                    "_base",
                    f"{new_name}.json",
                )
                makedirs(dirname(new_path), exist_ok=True)
                with open(new_path, "w") as f:
                    f.write("{}\n")
            case "Board base...", "boardgen directory":
                new_name = self.AskFileName()
                new_path = join(
                    self.core.dir_base,
                    "boards",
                    "_base",
                    f"{new_name}.json",
                )
                makedirs(dirname(new_path), exist_ok=True)
                with open(new_path, "w") as f:
                    f.write("{}\n")

            case "Template...", "LibreTiny directory":
                boards_dir = join(self.lvm.path(), "boards")
                self.CreateNewFile("templates", boards_dir, INIT_TEMPLATE)
            case "Template...", "boardgen directory":
                self.CreateNewFile("templates", self.core.dir_base, INIT_TEMPLATE)

            case "Shape...", "LibreTiny directory":
                boards_dir = join(self.lvm.path(), "boards")
                self.CreateNewFile("shapes", boards_dir, INIT_SHAPE)
            case "Shape...", "boardgen directory":
                self.CreateNewFile("shapes", self.core.dir_base, INIT_SHAPE)

            case _:
                if self.modified:
                    wx.MessageBox("Please save the changes first", "Information")
                    return
                if not self.edit_item:
                    return
                draw_item = self.draw_item
                old_path = self.file_map.get(draw_item, None)
                if not old_path:
                    return
                path_dir = dirname(old_path)
                old_name = basename(old_path).rpartition(".")[0]
                match label:
                    case "Rename item":
                        new_name = self.AskFileName(old_name)
                        if not new_name:
                            return
                        new_path = join(path_dir, f"{new_name}.json")
                        rename(old_path, new_path)
                        self.ReloadLists()
                        self.draw_item = draw_item.replace(old_name, new_name)
                    case "Delete item":
                        choice = wx.MessageBox(
                            f"Do you really want to delete '{draw_item}'?",
                            "Are you sure?",
                            wx.YES_NO | wx.ICON_QUESTION,
                        )
                        if choice != wx.YES:
                            return
                        unlink(old_path)
                        self.ReloadLists()
                        self.draw_item = self.DrawItem.GetItems()[0]
                    case "Duplicate item":
                        new_name = self.AskFileName(old_name)
                        if not new_name or new_name == old_name:
                            return
                        new_path = join(path_dir, f"{new_name}.json")
                        copyfile(old_path, new_path)
                        self.ReloadLists()
                        self.draw_item = draw_item.replace(old_name, new_name)

    def UpdateEditType(self) -> None:
        prev_type = self.edit_type
        if not self.edit_item:
            return
        name, _ = self.edit_item
        path = self.edit_path

        self.edit_type = None
        is_shape = (
            name.startswith("templates")
            and (path.startswith("front") or path.startswith("back"))
            or name.startswith("boards")
            and (path.startswith("pcb.front") or path.startswith("pcb.back"))
            or name.startswith("shapes")
        )
        is_vars = "vars." in path

        if is_shape and path.endswith(".name"):
            self.edit_type = EditType.SHAPE
        elif is_shape and path.endswith(".preset"):
            self.edit_type = EditType.PRESET
        elif is_vars and path.endswith("_PRESET"):
            self.edit_type = EditType.PRESET
        elif is_vars and path.endswith("_COLOR"):
            self.edit_type = EditType.COLOR
        elif is_vars and ".PINTYPE" in path:
            self.edit_type = EditType.SHAPE
        elif path.endswith(".color"):
            self.edit_type = EditType.COLOR
        elif path.endswith(".lgrad.1") or path.endswith(".lgrad.3"):
            self.edit_type = EditType.COLOR
        elif path.endswith(".family"):
            self.edit_type = EditType.FAMILY
        elif path.startswith("connectivity"):
            self.edit_type = EditType.CONNECTIVITY
        elif path.startswith("pcb.pinout.") or path.startswith("pcb.ic."):
            self.edit_type = EditType.ROLE
        elif path.startswith("pcb.pinout_hidden"):
            self.edit_type = EditType.ROLE_HIDDEN
        elif path.startswith("flash"):
            self.edit_type = EditType.FLASH
        elif path.startswith("pcb.templates"):
            self.edit_type = EditType.TEMPLATE
        elif is_shape:
            self.edit_type = EditType.SHAPE_ADD
        elif name.startswith("boards") and "_base" not in name:
            self.edit_type = EditType.BASE

        if self.edit_type:
            self.Choose.Enable(True)
            self.Choose.SetLabel(self.edit_type.name.replace("_", " ").title() + "...")
        else:
            self.Choose.Enable(False)
        if self.edit_type != prev_type:
            debug(f"Chooser type: {self.edit_type}")

    @on_event
    def OnChooseClick(self) -> None:
        if self.edit_errors:
            wx.MessageBox(
                "Can't do this, the JSON has syntax errors",
                "Error",
                wx.OK | wx.CENTRE | wx.ICON_ERROR,
            )
            return
        data = self.data
        if data is None:
            return
        path = self.edit_path

        if self.edit_type == EditType.SHAPE_ADD:
            # special case for adding shapes - find the shape list
            parts = path.split(".")
            new_path = []
            for part in parts:
                if part.isnumeric():
                    break
                new_path.append(part)
            if new_path:
                path = ".".join(new_path)
            elif isinstance(data, list):
                # path starts with numeric value - data object is a list
                path = parts[0]

        walk = jsonwalk(data, path)
        if not walk:
            return
        parent, key = walk
        this_value = parent[key]
        this_dict = this_value if isinstance(this_value, dict) else parent
        this_list = parent if isinstance(parent, list) else this_value
        new_value: str | int | None = None
        new_dict: dict | None = None
        new_list: list | None = None

        match self.edit_type:
            # multiple choice
            case EditType.BASE:
                items = self.core.list_json("boards", recursive=True)
                items = [i[6:] for i in items if i.startswith("_base/")]
                items = sorted(items, key=lambda i: "_" + i if "/" not in i else i)
                data["_base"] = self.AskMultipleChoice(
                    title="Choose board base",
                    items=items,
                    selected=data.get("_base", []),
                    sort=False,
                )
            case EditType.TEMPLATE:
                items = self.core.list_json("templates")
                items = sorted(items)
                new_list = self.AskMultipleChoice(
                    title="Choose PCB templates",
                    items=items,
                    selected=this_list,
                )
            case EditType.CONNECTIVITY:
                new_list = self.AskMultipleChoice(
                    title="Choose module connectivity options",
                    items=["wifi", "ble"],
                    selected=this_list,
                    sort=False,
                )
            case EditType.ROLE_HIDDEN:
                roles = self.AskMultipleChoice(
                    title="Choose hidden pinout roles",
                    items=[i.name for i in self.core.roles.keys()],
                    selected=this_value.split(","),
                    anchor=self.Choose,
                )
                new_value = ",".join(roles)
            # single choice
            case EditType.SHAPE:
                items = self.core.list_json("shapes")
                items = sorted(items)
                new_value = self.AskSingleChoice(
                    title="Choose shape",
                    items=items,
                    selected=this_value,
                    anchor=self.Choose,
                )
            case EditType.SHAPE_ADD:
                shape_type = self.AskSingleChoice(
                    title="Choose shape type to add",
                    items=[i.name.lower() for i in self.core.shape_ctors.keys()],
                    anchor=self.Choose,
                )
                default = {"type": shape_type}
                shape = SHAPE_DEFAULTS.get(shape_type, default)
                this_list.append(deepcopy(shape))
            case EditType.PRESET:
                new_value = self.AskSingleChoice(
                    title="Choose shape preset",
                    items=list(self.core.presets.keys()),
                    selected=this_value,
                    anchor=self.Choose,
                )
            case EditType.FAMILY:
                new_value = self.AskSingleChoice(
                    title="Choose chip family",
                    items=[f.short_name for f in Family.get_all() if f.is_chip],
                    selected=this_value,
                    sort=False,
                    anchor=self.Choose,
                )
            case EditType.ROLE:
                roles = set(i.name for i in self.core.roles.keys())
                roles.update(ROLE_DEFAULTS.keys())
                roles = sorted(roles)
                roles.remove("IC")
                roles.insert(0, "IC")
                role = self.AskSingleChoice(
                    title="Choose pin role to add",
                    items=roles,
                    sort=False,
                    anchor=self.Choose,
                )
                if not role:
                    return
                role_value = ROLE_DEFAULTS.get(role, "")
                if (
                    role == "IC"
                    and self.draw_item.startswith("boards/")
                    and self.draw_object
                    and self.draw_object.pcb
                    and self.draw_object.pcb.ic
                ):
                    ic_pins = {}
                    for ic_pin, role_map in self.draw_object.pcb.ic.items():
                        pin_roles = []
                        for role_type, functions in role_map.items():
                            if role_type == RoleType.IO:
                                continue
                            role_obj = self.core.role(role_type)
                            if not role_obj:
                                continue
                            pin_roles += role_obj.format(functions, long=False)
                        ic_pins[" / ".join(pin_roles)] = ic_pin
                    ic_role = self.AskSingleChoice(
                        title="Choose pin to add",
                        items=list(ic_pins.keys()),
                        anchor=self.Choose,
                    )
                    if not ic_role:
                        return
                    role_value = ic_pins[ic_role]
                ex_roles = ["PWR", "GND", "NC"]
                for ex_role in ex_roles:
                    if role == ex_role:
                        this_dict.clear()
                    else:
                        this_dict.pop(ex_role, None)
                this_dict[role] = role_value
            case EditType.FLASH:
                region = self.AskSingleChoice(
                    title="Choose flash region to add",
                    items=list(self.core.flash.keys()),
                    anchor=self.Choose,
                )
                if region not in this_dict:
                    this_dict[region] = "0x000000+0x0000"
            # other
            case EditType.COLOR:
                cdata = wx.ColourData()
                cdata.SetChooseFull(True)
                cdata.SetChooseAlpha(False)
                cdata.SetColour(this_value)
                dialog = wx.ColourDialog(
                    self,
                    data=cdata,
                )
                if dialog.ShowModal() == wx.ID_OK:
                    cdata = dialog.GetColourData()
                    color: wx.Colour = cdata.GetColour()
                    new_value = color.GetAsString(wx.C2S_NAME | wx.C2S_HTML_SYNTAX)
                    new_value = new_value.replace(" ", "")
                dialog.Destroy()

        if new_value and new_value != this_value and isinstance(parent, (dict, list)):
            parent[key] = new_value
        elif new_dict and new_dict is not this_dict:
            this_dict.clear()
            this_dict.update(new_dict)
        elif new_list and new_list is not this_list:
            this_list.clear()
            this_list += new_list
        self.UpdateEditItem()
        self.UpdateDrawItem()
        self.MarkModified()

    def AskFileName(self, value: str = "") -> str | None:
        dialog = wx.TextEntryDialog(
            self,
            message="Enter the new file name (without .json):",
            caption="File name",
            value=value,
        )
        if dialog.ShowModal() != wx.ID_OK:
            dialog.Destroy()
            return
        value = dialog.GetValue().strip()
        dialog.Destroy()
        return value

    def AskMultipleChoice(
        self,
        title: str,
        items: list[str],
        selected: list[str],
        sort: bool = True,
        anchor: wx.Window = None,
    ) -> list[str]:
        if sort:
            items = sorted(items)
        choice = []
        missing = []
        for item in selected[::-1]:
            if item not in items:
                missing.insert(0, item)
                continue
            items.remove(item)
            items.insert(0, item)

        if anchor:
            menu = wx.Menu()
            for item in items:
                menu_item: wx.MenuItem = menu.AppendCheckItem(wx.ID_ANY, item)
                menu_item.Check(item in selected)
            if anchor.PopupMenu(menu, 0, 25):
                choice = [
                    items[idx]
                    for idx, item in enumerate(menu.GetMenuItems())
                    if item.IsChecked()
                ]
            menu.Destroy()
        else:
            dialog = wx.MultiChoiceDialog(
                self,
                message=title,
                caption="Choose items",
                choices=items,
            )
            if selected:
                dialog.SetSelections([items.index(i) for i in selected if i in items])
            if dialog.ShowModal() == wx.ID_OK:
                choice = [items[i] for i in dialog.GetSelections()]
            dialog.Destroy()

        return choice + missing if choice else selected

    def AskSingleChoice(
        self,
        title: str,
        items: list[str],
        selected: str | None = None,
        sort: bool = True,
        anchor: wx.Window = None,
    ) -> str:
        if sort:
            items = sorted(items)
        choice = None

        if anchor:
            menu = wx.Menu()
            if selected:
                for item in items:
                    menu_item: wx.MenuItem = menu.AppendRadioItem(wx.ID_ANY, item)
                    menu_item.Check(item == selected)
                if anchor.PopupMenu(menu, 0, 25):
                    choice = next(
                        items[idx]
                        for idx, item in enumerate(menu.GetMenuItems())
                        if item.IsChecked()
                    )
            else:
                for item in items:
                    menu.Append(wx.ID_ANY, item)
                selection = anchor.GetPopupMenuSelectionFromUser(menu, 0, 25)
                if selection != -1:
                    for i, item in enumerate(items):
                        if menu.FindItemByPosition(i).GetId() == selection:
                            choice = item
                            break
            menu.Destroy()
        else:
            dialog = wx.SingleChoiceDialog(
                self,
                message=title,
                caption="Choose items",
                choices=items,
            )
            if selected in items:
                dialog.SetSelection(items.index(selected))
            if dialog.ShowModal() == wx.ID_OK:
                choice = (
                    items[dialog.GetSelection()]
                    if dialog.GetSelection() != -1
                    else None
                )
            dialog.Destroy()

        return choice or selected

    def CreateNewFile(self, item_type: str, directory: str, value: dict | list) -> None:
        name = self.AskFileName()
        if not name:
            return

        if value is INIT_BOARD:
            value["build"]["variant"] = name
            value["name"] = name
        if value is INIT_TEMPLATE:
            value["name"] = name
            value["title"] = name

        makedirs(join(directory, item_type), exist_ok=True)
        with open(join(directory, item_type, f"{name}.json"), "w") as f:
            json.dump(value, f, indent="\t")
            f.write("\n")
        self.ReloadLists()
        self.draw_item = f"{item_type}/{name}"
        self.Svg.ClearSvg()
        self.UpdateDrawItem()

    def SetError(self, e: Exception | None) -> None:
        if not e:
            self.Error.Clear()
        else:
            self.Error.SetValue(f"{type(e).__name__}: {e}")

    def MarkModified(self) -> None:
        if not self.edit_item:
            return
        file, obj = self.edit_item
        if file in self.modified:
            return
        self.modified[file] = obj
        self.Modified.SetLabel(f"Modified: {len(self.modified)}")
        self.Save.Enable(True)
        self.Discard.Enable(True)
        self.Revert.Enable(True)

    def MarkUnmodified(self) -> None:
        if not self.edit_item:
            return
        file, obj = self.edit_item
        if file not in self.modified:
            return
        self.modified.pop(file)
        self.Modified.SetLabel(f"Modified: {len(self.modified)}")
        self.Save.Enable(len(self.modified) > 0)
        self.Discard.Enable(len(self.modified) > 0)
        self.Revert.Enable(False)

    @property
    def user_vars(self) -> dict[str, str]:
        lines: list[str] = self.Vars.GetValue().splitlines()
        data = {}
        for line in lines:
            key, _, value = line.strip().partition("=")
            data[key] = value
        return data

    @user_vars.setter
    def user_vars(self, value: dict[str, str]) -> None:
        self.Vars.SetValue("\n".join(f"{k}={v}" for k, v in value.items()))

    @property
    def vars(self) -> dict[str, str]:
        data = self.user_vars
        data |= self._vars
        return data

    @vars.setter
    def vars(self, value: dict[str, str]) -> None:
        self._vars.clear()
        self._vars.update(value)

    @property
    def draw_item(self) -> str | None:
        return self.draw_items.get(self.DrawItem.GetStringSelection(), None)

    @draw_item.setter
    def draw_item(self, value: str) -> None:
        for key, item in self.draw_items.items():
            if key == value or item == value:
                self.DrawItem.SetStringSelection(key)
                self.UpdateDrawItem()
                return

    @property
    def edit_item(self) -> tuple[str, dict] | None:
        return self.edit_items.get(self.EditItem.GetStringSelection(), None)

    @edit_item.setter
    def edit_item(self, value: str) -> None:
        for key, item in self.edit_items.items():
            if key == value or item[1] == value == value:
                self.EditItem.SetStringSelection(key)
                self.UpdateEditItem()
                return

    @property
    def data(self) -> dict | None:
        item = self.edit_item
        return item and item[1]

    @data.setter
    def data(self, value: dict) -> None:
        data = self.data
        if data is None:
            return
        debug(f"Data change, id={id(data)}")
        data.clear()
        if isinstance(data, dict):
            data.update(value)
        if isinstance(data, list):
            data += value
